(defstruct circuit (gates nil :type list) (unresolved nil :type list))
(defstruct gate (pins nil :type list) (num 0 :type fixnum))
(defstruct pin (sym 'N/C :type symbol) (val 0 :type fixnum) link)

(defvar *number* nil)

(defun deep-copy-pins (pins)
  (mapcar (lambda (x) (copy-pin x)) pins))

(defun deep-copy-gate (gate)
  (let ((copy (copy-gate gate)))
    (setf (gate-pins copy) (deep-copy-pins (gate-pins copy)))
    copy))

(defun copy-circuit-gates (gates)
  (when gates
    (cons (deep-copy-gate (first gates))
	  (copy-circuit-gates (rest gates)))))

(defun fix-pin-links (gates)    
  (dolist (gate gates gates)
    (dolist (pin (gate-pins gate))
      (when (pin-link pin)
	(setf (pin-link pin) (elt gates (gate-num (pin-link pin))))))))

(defun pick-unresolved (gates circuit)
  (mapcar (lambda (x) (elt gates (gate-num x)))
	  (circuit-unresolved circuit)))

(defun deep-copy-circuit (circuit)
  (let ((copy (copy-circuit-gates (circuit-gates circuit))))
    (make-circuit :unresolved (pick-unresolved copy circuit)
		  :gates (fix-pin-links copy))))

(defun create-pin (sym)
  (make-pin :sym sym))

(defun create-gate (pins)
  (prog1
      (make-gate :pins (mapcar #'create-pin pins) :num *number*)
    (incf *number*)))

(defun set-link-in-gate (pin gate link)  
  (declare (optimize (speed 3) (safety 0)))
  (dolist (i (gate-pins gate))
    (when (and (eq (pin-sym pin) (pin-sym i))
	       (null (pin-link pin))
	       (null (pin-link i)))
      (setf (pin-link pin) gate)
      (setf (pin-link i) link))))

(defun set-gate-links (source-gate gate item)
  (declare (optimize (speed 3) (safety 0)))
  (mapc (lambda (x) (set-link-in-gate x gate item)) (gate-pins source-gate)))

(defun check-pin-sym (x y)
  (eq x (pin-sym y)))

(defun set-self-links-in-pin-list (gate head tail)
  (when tail
    (let ((item (member (pin-sym head) tail :test #'check-pin-sym)))
      (when (and item (eq (pin-sym head) (pin-sym (first item))))
	(when (and (null (pin-link (first item)))
		   (null (pin-link head)))
	  (setf (pin-link (first item)) gate)
	  (setf (pin-link head) gate)))
      (set-self-links-in-pin-list gate (first tail) (rest tail)))))

(defun set-self-links (gate)
  (declare (optimize (speed 3) (safety 0)))
  (let ((pins (gate-pins gate)))
    (set-self-links-in-pin-list gate (first pins) (rest pins))))

(defun set-circuit-links (gate circuit)
  (declare (optimize (speed 3) (safety 0)))
  (mapc (lambda (x) (set-gate-links gate x gate)) circuit)
  (set-self-links gate))

(defun set-in-gate (symbol gate fn)
  (dolist (i (gate-pins gate))
    (when (eq symbol (pin-sym i))
      (funcall fn i))))

(defun set-in-circuit (symbol circuit fn)
  (mapc (lambda (gate) (set-in-gate symbol gate fn)) (circuit-gates circuit)))

(defun set-value-in-circuit (symbol circuit value)
  (set-in-circuit symbol circuit (lambda (pin) (setf (pin-val pin) value))))

(defun set-sym-in-circuit (symbol circuit sym)
  (set-in-circuit 
   symbol circuit 
   (lambda (pin)
     (when (null (pin-link pin))
       (setf (pin-sym pin) sym)))))

(defun link-circuit-return-nil (circuit)
  (declare (optimize (speed 3) (safety 0)))
  (when circuit
    (set-circuit-links (first circuit) (rest circuit))
    (link-circuit-return-nil (rest circuit))))

(defmacro l-in (pins) `(first ,pins))
(defmacro l-out (pins) `(second ,pins))
(defmacro r-in (pins) `(third ,pins))
(defmacro r-out (pins) `(fourth ,pins))

(defun is-resolved (gate)
  (not (member nil (gate-pins gate) :test (lambda (x y) (eq x (pin-link y))))))

(defun link-circuit (circuit)
  (let ((unresolved (circuit-unresolved circuit)))
    (link-circuit-return-nil unresolved)
    (setf (circuit-unresolved circuit) (remove-if #'is-resolved unresolved))
    circuit))

(defun create-circuit (&rest rest)
  (let* ((*number* 0)
	 (gates (mapcar #'create-gate rest)))
    (link-circuit (make-circuit :gates gates :unresolved gates))))

(defun inc-numbering (circuit number)
  (mapc (lambda (x) (incf (gate-num x) number)) (circuit-gates circuit))
  circuit)

(defun link-circuits (c1 c2)
  (inc-numbering c2 (length (circuit-gates c1)))
  (link-circuit
   (make-circuit
    :gates (append (circuit-gates c1) (circuit-gates c2))
    :unresolved (append (circuit-unresolved c1) (circuit-unresolved c2)))))

(defun safe-gate-num (gate)
  (when gate (gate-num gate)))

(defun format-pin (pin)
  (format nil "~A" (pin-val pin)))

(defun print-gate (gate)
  (let ((pins (gate-pins gate)))
    (format t "~3,'0d: (~A;~A) -> (~A;~A)~%"
	    (gate-num gate)
	    (format-pin (l-in pins))
	    (format-pin (r-in pins))
	    (format-pin (l-out pins))
	    (format-pin (r-out pins)))))

(defun print-circuit (circuit)
  (dolist (i (circuit-gates circuit))
    (print-gate i)))

(defun propagate-to-pin (pin1 pin2)
  (declare (optimize (speed 3) (safety 0)))
  (when (eq (pin-sym pin1) (pin-sym pin2))
    (setf (pin-val pin1) (pin-val pin2))))

(defun propagate-value (pin)
  (declare (optimize (speed 3) (safety 0)))
  (when (pin-link pin)
    (let ((pins (gate-pins (pin-link pin))))
      (propagate-to-pin (l-in pins) pin)
      (propagate-to-pin (r-in pins) pin))))

(defun propagate-outputs (gate)
  (declare (optimize (speed 3) (safety 0)))
  (propagate-value (l-out (gate-pins gate)))
  (propagate-value (r-out (gate-pins gate))))

(defun propagate-signals (gate)
  (declare (optimize (speed 3) (safety 0)))
  (let* ((pins (gate-pins gate))
	 (l-in-v (pin-val (l-in pins)))
	 (r-in-v (pin-val (r-in pins))))
    (setf (pin-val (l-out pins)) (mod (- l-in-v r-in-v) 3))
    (setf (pin-val (r-out pins)) (mod (1- (the fixnum (* l-in-v r-in-v))) 3))
    ; (print-gate gate)
    (propagate-outputs gate)))

(defun char-to-num (c)
  (case c
    (#\0 0)
    (#\1 1)
    (#\2 2)))

(defun num-elt (s i)
  (char-to-num (elt s i)))

(defun search-symbol (circuit sym fn)
  (dolist (gate (circuit-gates circuit))
    (dolist (pin (gate-pins gate))
      (when (eq sym (pin-sym pin))
	(funcall fn gate pin)))))

(defvar *output* nil)

(defun get-sym-pin (circuit out-sym)
  (search-symbol
   circuit out-sym
   (lambda (gate pin)
     (declare (ignore gate))
     (return-from get-sym-pin pin))))
 
(defun evaluate-circuit (circuit in-sym out-sym string)
  (let (return-string return-pin input-pin)
    (setf return-pin (get-sym-pin circuit out-sym))
    (setf input-pin (get-sym-pin circuit in-sym))
    (with-output-to-string (*output*)
      (dotimes (i (length string))
	(setf (pin-val input-pin) (num-elt string i))
	(mapc #'propagate-signals (circuit-gates circuit))
	(format *output* "~A" (pin-val return-pin)))
      (setf return-string(get-output-stream-string *output*)))
    return-string))

(defun pin-side (pin gate type)
  (case type
    (:i (if (eq (pin-sym pin) (pin-sym (r-out (gate-pins gate)))) "R" "L"))
    (:o (if (eq (pin-sym pin) (pin-sym (r-in (gate-pins gate)))) "R" "L"))))

(defvar *exceptions* nil)

(defun print-internal-pin (pin &key gate type)
  (if (and (not gate) (member (pin-sym pin) *exceptions*))
      (format t "X")
      (let ((gate (or (pin-link pin) gate)))
	(if (null gate)
	    (format t "??")
	    (format t "~A~A" (gate-num gate) (pin-side pin gate type))))))

(defun print-external-pin (type)
  (lambda (gate pin)
    (print-internal-pin pin :gate gate :type type)))

(defun compile-gate (circuit in-sym out-sym)
  (when circuit
    (let ((pins (gate-pins (first circuit))))      
      (print-internal-pin (l-in pins) :type :i)
      (print-internal-pin (r-in pins) :type :i)
      (format t "0#")
      (print-internal-pin (l-out pins) :type :o)
      (print-internal-pin (r-out pins) :type :o)
      (format t "~A~%" (if (rest circuit) "," ":"))
      (compile-gate (rest circuit) in-sym out-sym))))

(defun compile-circuit (circuit in-sym out-sym)
  (let ((*exceptions* (list in-sym out-sym)))
    (search-symbol circuit in-sym (print-external-pin :o))
    (format t ":~%")
    (compile-gate (circuit-gates circuit) in-sym out-sym)
    (search-symbol circuit out-sym (print-external-pin :i))
    (format t "~%")))

(defun test ()
  (create-circuit
   '(A C
     X Y)
   '(C B
     D E)
   '(B D
     E H)
   '(K K
     H A)))

(defun do-test ()
  ; 21120221002022020
  (format t "\"~A\"~%" (evaluate-circuit (test) 'X 'Y "01202101210201202"))
  (compile-circuit (test)  'X 'Y)
  (print-circuit (test)))

(defun G0->0 ()
  (create-circuit
   '(in out
     up down)))

(defun G0->1 ()
  (create-circuit
   '(A A
     B C)
   '(in D
     up B)
   '(D out
     C down)))

(defun G0->2 ()
  (create-circuit
   '(A A
     B C)
   '(in D
     up B)
   '(C out
     D down)))
   
(defun G1->0 ()
  (create-circuit
   '(B X1
     A X2)
   '(X1 C
     X2 A)
   '(in D
     up B)
   '(D out
     C down)))

(defun G1->1 ()
  (G0->0))

(defun G1->2 ()
  (create-circuit
   '(up out
     in down)))

(defun G2->0 ()
  (G0->1))

(defun G2->1 ()
  (G1->2))

(defun G2->2 ()
  (G0->0))

(defparameter *table* (vector (vector #'G0->0 #'G0->1 #'G0->2)
			      (vector #'G1->0 #'G1->1 #'G1->2)
			      (vector #'G2->0 #'G2->1 #'G2->2)))

(defparameter *the-list* (list #'G0->0 #'G1->2 #'G0->1 #'G0->2 #'G1->0))

(defun choose-first (input output)
  (let ((start (funcall (aref (aref *table* input) output))))
    (set-sym-in-circuit 'in start 'INPUT)
    (set-sym-in-circuit 'out start 'OUTPUT)
    start))  

(defun reset-circuit (circuit)
  (dolist (gate (circuit-gates circuit))
    (dolist (pin (gate-pins gate))
      (setf (pin-val pin) 0))))

(defun choose-element (input output circuit)
  (dolist (fn *the-list*)
    (let* ((candidate (funcall fn))
	   (product (link-circuits (deep-copy-circuit circuit) candidate)))
      ;; (compile-circuit product 'INPUT 'OUTPUT)
      (when (equal output (evaluate-circuit product 'INPUT 'OUTPUT input))
	(reset-circuit product)
	(return-from choose-element product))))
  (error "sucks compleatly!~%"))

(defun add-to-circuit (from-str to-str i circuit)
  (if (= 0 i)
      (choose-first (num-elt from-str i) (num-elt to-str i))
      (choose-element (subseq from-str 0 (1+ i))
		      (subseq to-str 0 (1+ i))
		      circuit)))

(defun build-circuit (from-str to-str)
  (let (circuit)
    (dotimes (i (min (length from-str) (length to-str)))
      (setf circuit (add-to-circuit from-str to-str i circuit))
      (set-sym-in-circuit 'up circuit 'out)
      (set-sym-in-circuit 'down circuit 'in))
    (set-sym-in-circuit 'out circuit 'xxx)
    (set-sym-in-circuit 'in circuit 'xxx)
    (link-circuit circuit)
    circuit))

(defun ultimate ()
  (build-circuit "01202101210201202" "11021210112101221"))

(defparameter *prefix* "11021210112101221")

(defun generic (string)
  (build-circuit 
   (make-string (+ (length string) (length *prefix*)) :initial-element #\0)
   (concatenate 'simple-string *prefix* string)))
		
(defun compile-generic (string)
  (compile-circuit 
   (link-circuits (nullatron) (generic string))
   'SUPER-INPUT 'OUTPUT))

(defun nullatron ()
  (create-circuit
   '(SUPER-INPUT L0
     L2 R0)
   '(R0 L1
     L0 R1)
   '(L1 L2
     R1 R2)
   '(L5 L3
     R2 R3)
   '(R5 L4
     L4 R4)
   '(R3 L5
     R4 R5)
   '(L3 L6
     L8 R6)
   '(R6 L7
     L6 R7)
   '(L7 L8
     R7 INPUT)))

(defun the-test ()
  (compile-generic "101010100101201201200122010201201020012012001020120102010201200120102021020021210101010010120120120012201020120102001201200102012010201020120012010202102002120010101010010120120120012201020120102001201200102012010201020120012010202102002121"))

;(defun profile-test ()
;  (swank::toggle-profile-fdefinition "the-test")
;  (swank::toggle-profile-fdefinition "link-circuit")
;  (swank::toggle-profile-fdefinition "set-link-in-gate")
;  (swank::toggle-profile-fdefinition "set-self-links")
;  (swank::toggle-profile-fdefinition "propagate-signals")
;  (the-test)
;  (swank::profile-report)
;  (swank::profile-reset)
;  (swank::unprofile-all))
